#!/usr/bin/env python3
import numpy as np
import os
import sys
from scipy.stats import norm
import time
import subprocess
import pandas as pd

##SOME CONSTANTS##############################################
epsilon0 = 8.854187817e-12   #F/m
hbar = 6.582119514e-16       #eV s
hbar2 = 1.054571800e-34      #J s
mass = 9.10938356e-31        #kg
c = 299792458                #m/s
e = 1.60217662e-19           #C
kb = 8.6173303e-5            #eV/K
amu = 1.660539040e-27        #kg
pi = np.pi
###############################################################

##ERROR FUNCTION###############################################
def fatal_error(msg):
    print(msg)
    sys.exit()
###############################################################

##GETS FREQUENCIES AND REDUCED MASSES##########################
def pega_freq(freqlog):
    F, M = [], []
    with open(freqlog, 'r') as f:
        for line in f:
            if "Frequencies --" in line:
                line = line.split()
                for j in range(2,len(line)):
                    if float(line[j]) in F:
                        pass
                    F.append(float(line[j]))
            elif "Red. masses --" in line:
                line = line.split()
                for j in range(3,len(line)):
                    M.append(float(line[j]))
            elif 'Thermochemistry' in line:
                break        
    #conversion in angular frequency
    F = np.array(F)*(c*100*2*pi) 
    try:
        f = F[0]
    except:
        fatal_error("No frequencies in the log file! Goodbye!")
    #conversion from amu to kg
    M = np.asarray(M)*amu
    return F, M
###############################################################

##GETS ATOMS AND LAST GEOMETRY IN FILE#########################
def pega_geom(freqlog):
    if ".log" in freqlog:
        status = 0
        busca = "orientation:"
        n = -1
        with open(freqlog, 'r') as f:
            for line in f:
                if busca in line and 'Dipole' not in line:
                    n = 0
                    G = np.zeros((1,3))
                    atomos = []
                elif n >= 0 and n < 4:
                    n += 1
                elif n >= 4 and "---------------------------------------------------------------------" not in line:    
                    line = line.split()
                    NG = []
                    for j in range(3,len(line)):
                        NG.append(float(line[j]))
                    atomos.append(line[1])
                    G = np.vstack((G,NG))       
                    n += 1  
                elif "---------------------------------------------------------------------" in line and n>1:
                    n = -1       
    else:
        G = np.zeros((1,3))
        atomos = []
        with open(freqlog, 'r') as f:
            for line in f:
                line = line.split()
                try:
                    vetor = np.array([float(line[1]),float(line[2]), float(line[3])])
                    atomos.append(line[0])
                    G = np.vstack((G,vetor))
                except:
                    pass
    G = G[1:,:]                 
    return G, atomos
###############################################################

##SAVES OPT GEOMETRY###########################################
def salva_geom(G,atomos):
    atomos = np.array([atomos]).astype(float)
    atomos = atomos.T
    G = np.hstack((atomos,G))
    np.savetxt('opt_geom.lx', G, delimiter='\t', fmt=['%1.1u','%+1.5f','%+1.5f','%+1.5f'])
    print("The optimized geometry that is used is saved in the opt_geom.lx file!")
###############################################################

##GETS NORMAL COORDINATES IN HIGH PRECISION####################
def pega_modosHP(G, freqlog):
    F, M = pega_freq(freqlog)
    n = -1
    num_atom = np.shape(G)[0]
    NC = np.zeros((3*num_atom,1))
    with open(freqlog, 'r') as f:
        for line in f:
            if n == 0:
                line = line.split()[3:]
                C = np.asarray([float(i) for i in line])
                n += 1
            elif n < 0 or n > 3*num_atom:
                if "Coord Atom Element:" in line:
                    n = 0
            elif n > 0 and n < 3*num_atom:
                line = line.split()[3:]
                line = np.asarray([float(i) for i in line])
                C = np.vstack((C,line))
                n += 1  
            elif n == 3*num_atom:
                NC = np.hstack((NC,C))
                n += 1
    NC = NC[:,1:]
    MM = np.zeros((1,len(F)))
    M = np.expand_dims(M,axis=0)
    for _ in range(0,3*num_atom):
        MM = np.vstack((MM,M))
    M = MM[1:,:]
    return NC
###############################################################

##GETS NORMAL COORDINATES IN REGULAR PRECISION#################
def pega_modosLP(G,freqlog):
    F, M = pega_freq(freqlog)
    C = []
    n = -1
    num_atom = np.shape(G)[0]
    with open(freqlog, 'r') as f:
        for line in f:
            if n < 0 or n >= num_atom:
                if "Atom  AN      X      Y      Z" in line:
                    n = 0
                else:
                    pass
            elif n >= 0 and n < num_atom:
                line = line.split()
                for j in range(2,len(line)):
                    C.append(float(line[j]))
                n += 1  
                
    num_modos = len(F)
    
    l = 0
    p = 0
    NNC = np.zeros((num_atom,1))
    while l < num_modos:
        NC = np.zeros((1,3))
        k =0
        while k < num_atom:     
            B = np.asarray(C[3*(l+3*k)+p:3*(l+3*k)+3+p])
            NC = np.vstack((NC,B))
            k += 1      
        NNC = np.hstack((NNC,NC[1:,:]))
        l += 1
        if l%3 == 0 and l != 0:
            p = p + (num_atom-1)*9  
    NNC = NNC[:,1:] #matriz com as coordenadas normais de cada modo
    D = np.zeros((3*num_atom,1))
    for i in range(0,len(F)):
        normal = NNC[:,3*i:3*i+3].flatten()
        normal = np.expand_dims(normal,axis=1)
        D = np.hstack((D,normal))
    D = D[:,1:]
    MM = np.zeros((1,len(F)))
    M = np.expand_dims(M,axis=0)
    for i in range(0,3*num_atom):
        MM = np.vstack((MM,M))
    M = MM[1:,:]
    return D
###############################################################

##DETECTS WHETHER HIGH PRECISION IS USED#######################
def pega_modos(G,freqlog):
    x = 'LP'
    with open(freqlog, 'r') as f:
        for line in f:
            if "Coord Atom Element:" in line:
                x = 'HP'
                break
    if x == 'LP':
        return pega_modosLP(G,freqlog)
    else:
        return pega_modosHP(G,freqlog)
###############################################################

##WRITES ATOMS AND XYZ COORDS TO FILE##########################
def write_input(atomos,G,header,bottom,file):
    with open(file, 'w') as f:
        f.write(header)
        for i in range(0,len(atomos)):
            texto = "{:2s}  {:.14f}  {:.14f}  {:.14f}\n".format(atomos[i],G[i,0],G[i,1],G[i,2])
            f.write(texto)
        f.write("\n"+bottom+'\n')
###############################################################

##CHECKS FOR EXISTING GEOMETRIES###############################
def start_counter():
    files = [file for file in os.listdir('Geometries') if ".com" in file and "Geometr" in file]
    return len(files)
###############################################################

##SAMPLES GEOMETRIES###########################################
def sample_geometries(freqlog,num_geoms,T, limit=np.inf):
    G, atomos = pega_geom(freqlog)
    F, M      = pega_freq(freqlog)
    F[F < 0] *= -1
    NNC       = pega_modos(G,freqlog)
    mask = F < limit*(c*100*2*pi)
    F = F[mask]
    NNC = NNC[:,mask]
    num_atom  = np.shape(G)[0]
    A = np.zeros((3*num_atom,num_geoms))
    for i in range(0,len(F)):
        scale = np.sqrt(hbar2/(2*M[i]*F[i]*np.tanh(hbar*F[i]/(2*kb*T))))
        normal = norm(scale=scale,loc=0)
        #Displacements in  Å
        q = normal.rvs(size=num_geoms)*1e10
        try:
            numbers = np.hstack((numbers,q[:,np.newaxis]))
        except:
            numbers = q[:,np.newaxis]
        A += np.outer(NNC[:,i],q)
    for n in range(np.shape(A)[1]):
        A1 = np.reshape(A[:,n],(num_atom,3))
        try:
            Gfinal = np.hstack((Gfinal,A1 + G))
        except:
            Gfinal = A1 + G     
    numbers = np.round(numbers,4)
    return numbers, atomos, Gfinal
###############################################################

##MAKES ENSEMBLE###############################################
def make_ensemble(freqlog, num_geoms, T, header, bottom):
    try:
        os.mkdir('Geometries')
    except:
        pass        
    counter = start_counter()   
    print("\nGenerating geometries...\n")
    numbers, atomos, A = sample_geometries(freqlog,num_geoms,T)
    F, M      = pega_freq(freqlog)
    #convert numbers to dataframe
    numbers = pd.DataFrame(numbers,columns=[f"mode_{i+1}" for i in range(np.shape(numbers)[1])])
    #check if file exists
    if os.path.isfile(f'Magnitudes_{T:.0f}K_.lx'):
        data = pd.read_csv(f'Magnitudes_{T:.0f}K_.lx')
        # get only columns with mode_ in the name
        data = data.filter(regex='mode_')
        #remove nan values
        data = data.dropna()
        # join data and numbers on axis 0
        numbers = pd.concat([data,numbers],axis=0,ignore_index=True)
    # concatenate frequencies and masses to numbers
    numbers = pd.concat([pd.DataFrame(F,columns=['freq']),pd.DataFrame(M,columns=['mass']),numbers],axis=1)
    numbers.to_csv(f'Magnitudes_{T:.0f}K_.lx',index=False)
    for n in range(0,np.shape(A)[1],3):
        Gfinal = A[:,n:n+3]  
        write_input(atomos,Gfinal,header.replace("UUUUU",str((n+3)//3)),bottom.replace("UUUUU",str((n+3)//3)),"Geometries/Geometry-"+str((n+3)//3+counter)+"-.com")
        progress = 100*((n+3)//3)/num_geoms
        text = "{:2.1f}%".format(progress)
        print(' ', text, "of the geometries done.",end="\r", flush=True)
    print("\n\nDone! Ready to run.")   
################################################################
            
##COLLECTS RESULTS############################################## 
def gather_data(opc, tipo):
    files = [file for file in os.listdir('Geometries') if ".log" in file and "Geometr" in file ]    
    files = [i for i in files if 'Normal termination' in open('Geometries/'+i, 'r').read()]
    files = sorted(files, key=lambda file: float(file.split("-")[1])) 
    with open("Samples.lx", 'w') as f:
        for file in files:
            num = file.split("-")[1]
            broadening = opc
            numeros, energies, fs, scfs = [], [], [], []
            corrected, total_corrected = -1, -1
            with open('Geometries/'+file, 'r') as g:
                for line in g: 
                    if "Excited State" in line:
                        line = line.split()
                        numeros.append(line[2])
                        energies.append(line[4])
                        fs.append(line[8][2:])
                    elif "Corrected transition energy" in line:
                        line = line.split()
                        corrected = line[4]
                    elif "Total energy after correction" in line:
                        line = line.split()
                        total_corrected = 27.2114*float(line[5])
                    elif "SCF Done:" in line:
                        line = line.split()
                        scfs.append(27.2114*float(line[4]))
                if len(numeros) > 0:
                    f.write("Geometry "+num+":  Vertical transition (eV) Oscillator strength Broadening Factor (eV) \n")
                    if corrected != -1 and tipo == 'abs': #abspcm
                        f.write("Excited State {}\t{}\t{}\t{}\n".format(numeros[0],corrected, fs[0], broadening))
                    elif corrected != -1 and tipo == 'emi': #emipcm     
                        energy = total_corrected - scfs[-1]
                        f.write("Excited State {}\t{:.3f}\t{}\t{}\n".format(numeros[0],energy,fs[0],broadening))
                    elif corrected == -1 and tipo == 'emi':
                        f.write("Excited State {}\t{}\t{}\t{}\n".format(numeros[0],energies[0],fs[0],broadening))
                    else:
                        for i in range(len(energies)):
                            f.write("Excited State {}\t{}\t{}\t{}\n".format(numeros[i],energies[i],fs[i],broadening))
                    f.write("\n")   
############################################################### 


##NORMALIZED GAUSSIAN##########################################
def gauss(x,v,s):
    y =  (1/(np.sqrt(2*np.pi)*s))*np.exp(-0.5*((x-v)/s)**2)
    return y
###############################################################


##COMPUTES AVG TRANSITION DIPOLE MOMENT########################
def calc_tdm(O,V):
    #Energy terms converted to J
    term = e*(hbar2**2)/V
    dipoles = np.sqrt(3*term*O/(2*mass))
    #Conversion in au
    dipoles *= 1.179474389E29
    return np.mean(dipoles)
###############################################################

##PREVENTS OVERWRITING#########################################
def naming(arquivo):
    new_arquivo = arquivo
    if arquivo in os.listdir('.'):
        duplo = True
        vers = 2
        while duplo:
            new_arquivo = str(vers)+arquivo
            if new_arquivo in os.listdir('.'):
                vers += 1
            else:
                duplo = False
    return new_arquivo        
###############################################################

##CALCULATES FLUORESCENCE LIFETIME IN S########################
def calc_emi_rate(xd,yd,dyd):
    #Integrates the emission spectrum
    IntEmi = np.trapz(yd,xd)
    taxa   = (1/hbar)*IntEmi
    error  = (1/hbar)*np.sqrt(np.trapz((dyd**2),xd))
    return taxa, error 
###############################################################

##COMPUTES SPECTRA############################################# 
def spectra(tipo, num_ex, nr):
    if tipo == "abs":
        constante = (np.pi*(e**2)*hbar)/(2*nr*mass*c*epsilon0)*10**(20)
    elif tipo == 'emi':
        constante = ((nr**2)*(e**2)/(2*np.pi*hbar*mass*(c**3)*epsilon0))
    V, O, S = [], [], []
    N = 0
    with open("Samples.lx", 'r') as f:
        for line in f:
            if "Geometry" in line:
                N += 1
            elif "Excited State" in line and int(line.split()[2][:-1]) in num_ex:
                line = line.split()
                if float(line[3]) <= 0:
                    print('Ignoring geom with negative vertical transition!')
                    N -= 1
                else: 
                    V.append(float(line[3]))
                    O.append(float(line[4]))
                    S.append(float(line[5]))
    coms = start_counter()
    if len(V) == 0 or len(O) == 0:
        fatal_error("You need to run steps 1 and 2 first! Goodbye!")
    elif len(V) != coms*max(num_ex):
        print("Number of log files is less than the number of inputs. Something is not right! Computing the spectrum anyway...")
    V = np.array(V)
    O = np.array(O)
    S = np.array(S)
    if tipo == 'abs':
        espectro = (constante*O)
    else:
        espectro = (constante*(V**2)*O)
        tdm = calc_tdm(O,V)
    left = max(min(V)-3*max(S),0.001)
    right = max(V)+ 3*max(S)
    x  = np.linspace(left,right, int((right-left)/0.01))    
    if tipo == 'abs':
        arquivo = 'cross_section.lx'
        primeira = "{:8s} {:8s} {:8s}\n".format("#Energy(ev)", "cross_section(A^2)", "error")
    else:
        arquivo = 'differential_rate.lx'
        primeira = "{:4s} {:4s} {:4s} TDM={:.3f} au\n".format("#Energy(ev)", "diff_rate", "error",tdm)
    arquivo = naming(arquivo)
    y = espectro[:,np.newaxis]*gauss(x,V[:,np.newaxis],S[:,np.newaxis])
    mean_y =   np.sum(y,axis=0)/N 
    #Error estimate
    sigma  =   np.sqrt(np.sum((y-mean_y)**2,axis=0)/(N*(N-1))) 

    if tipo == 'emi':
        #Emission rate calculations
        mean_rate, error_rate = calc_emi_rate(x, mean_y,sigma) 
        segunda = '# Total Rate S1 -> S0: {:5.2e} +/- {:5.2e} s^-1\n'.format(mean_rate,error_rate)
    else:
        segunda = '# Absorption from State: S0\n'

    print(N, "geometries considered.")     
    with open(arquivo, 'w') as f:
        f.write(primeira)
        f.write(segunda)
        for i in range(0,len(x)):
            text = "{:.6f} {:.6e} {:.6e}\n".format(x[i],mean_y[i], sigma[i])
            f.write(text)
    print('Spectrum printed in the {} file'.format(arquivo))                
############################################################### 

##CHECKS THE FREQUENCY LOG'S LEVEL OF THEORY###################
def busca_input(freqlog):
    base = 'lalala'
    exc = ''
    header = ''
    nproc = '4'
    mem   = '1GB'
    scrf  = ''
    with open(freqlog, 'r') as f:
        search = False
        for line in f:
            if '%nproc' in line.lower():
                line = line.split('=')
                nproc = line[-1].replace('\n','')
            elif '%mem' in line.lower():
                line = line.split('=')
                mem = line[-1].replace('\n','')
            elif "#" in line and not search and header == '':
                search = True
                header += line.lstrip().replace('\n','')
            elif search and '----------' not in line:
                header += line.lstrip().replace('\n','')
            elif search and '----------' in line:
                search = False
                break
  
    if 'TDA' in header.upper():
        exc = 'tda'
        spec = 'EMISPCT'
    elif 'TD' in header.upper():
        exc = 'td'
        spec = 'EMISPCT'
    else:
        spec = 'ABSSPCT'

    if 'SCRF' in header.upper():
        new = header.split()
        for elem in new:
            if 'SCRF' in elem:
                scrf = elem
                break     
        
    header = header.split()
    base = ''
    for elem in header:
        if "/" in elem and 'IOP' not in elem.upper():
            base += elem.replace('#','')
        elif 'IOP' in elem.upper() and ('108' in elem or '107' in elem):
            base += ' '+elem
    return base, exc, nproc, mem, scrf, spec                
###############################################################

##CHECKS PROGRESS##############################################
def andamento():
    try:
        coms = [file for file in os.listdir("Geometries") if 'Geometr' in file and '.com' in file]
        logs = [file for file in os.listdir("Geometries") if 'Geometr' in file and '.log' in file]
        factor = 1
        with open('Geometries/'+coms[0], 'r') as f:
            for line in f:
                if 'Link1' in line:
                    factor = 2
        count = 0
        error = 0
        for file in logs:
            with open('Geometries/'+file, 'r') as f:
                for line in f:
                    if "Normal termination" in line:
                        count += 1
                    elif "Error termination" in line:
                        error += 1    
        print("\n\nThere are", int(count/factor), "successfully completed calculations out of", len(coms), "inputs")
        if error > 0:
            print("There are {} failed jobs. If you used option 2, check the nohup.out file for details.".format(error))                
        print(np.round(100*(count+error)/(factor*len(coms)),1), "% of the calculations have been run.")
    except:
        print('No files found! Check the folder!')                    
###############################################################


##FETCHES  FILES###############################################
def fetch_file(frase,ends):
    files = []
    for file in [i for i in os.listdir('.')]:
        for end in ends:
            if end in file:
                 files.append(file)
    if len(files) == 0:
        fatal_error("No {} file found. Goodbye!".format(frase))
    freqlog = 'nada0022'    
    for file in sorted(files):
        print("\n"+file)
        resp = input('Is this the {} file? y ou n?\n'.format(frase))
        if resp.lower() == 'y':
            freqlog = file
            break
    if freqlog == 'nada0022':
        fatal_error("No {} file found. Goodbye!".format(frase))
    return freqlog  
###############################################################  
   
##RUNS TASK MANAGER############################################
def batch():
    script   = fetch_file('batch script?',['.sh'])    
    num      = input("Number of jobs in each batch?\n")
    limite   = input("Maximum number of jobs to be submitted simultaneously?\n")
    gaussian = input('g16 or g09?\n')
    try:
        int(limite)
        int(num)
    except:
        fatal_error("These must be integers. Goodbye!")
    
    import subprocess
    folder = os.path.dirname(os.path.realpath(__file__)) 
    with open('limit.lx','w') as f:
        f.write(limite)
    subprocess.Popen(['nohup', 'python3', folder+'/batch_lx.py', script, num, gaussian, '&'])
###############################################################


##RUNS W TUNING################################################
def omega_tuning():
    geomlog = fetch_file('input or log',['.com','.log'])
    base, _, nproc, mem, _, _ = busca_input(geomlog)
    if 'IOP' in base.upper() and ('108' in base or '107' in base):
        base2 = base.split()
        for elem in base2:
            if '/' in elem:
                base = elem
                break
    omega1 = '0.1'
    passo  = '0.025'
    relax  = 'y'
    print('This is the configuration taken from the file:\n')
    print('Functional/basis: {}'.format(base))
    print('%nproc='+nproc)    
    print('%mem='+mem)
    print('Initial Omega: {} bohr^-1'.format(omega1))
    print('Step: {} bohr^-1'.format(passo))
    print('Optimize at each step: yes')
    change = input('Are you satisfied with these parameters? y or n?\n')
    if change == 'n':
        base   = default(base,"Functional/basis is {}. If ok, Enter. Otherwise, type functional/basis.\n".format(base))
        nproc  = default(nproc,'nproc={}. If ok, Enter. Otherwise, type it.\n'.format(nproc))
        mem    = default(mem,"mem={}. If ok, Enter. Otherwise, type it.\n".format(mem))
        omega1 = default(omega1,"Initial omega is {} bohr^-1. If ok, Enter. Otherwise, type it.\n".format(omega1))       
        passo  = default(passo,"Initial step is {} bohr^-1. If ok, Enter. Otherwise, type it.\n".format(passo))       
        relax  = default(relax,"Optimize at each step: yes. If ok, Enter. Otherwise, type n\n")
    
    script = fetch_file('batch script',['.sh'])    
    gaussian = input('g16 or g09?\n')
    import subprocess
    folder = os.path.dirname(os.path.realpath(__file__)) 
    with open('limit.lx','w') as f:
        f.write('Running')
    subprocess.Popen(['nohup', 'python3', folder+'/omega.py', geomlog, base, nproc, mem, omega1, passo, relax, script, gaussian, '&'])
###############################################################

##RUNS CONFORMATIONAL SEARCH###################################
def conformational():
    freqlog = fetch_file('frequency',['.log'])
    F, _ = pega_freq(freqlog)
    F_active = F[:40]
    T  = int(hbar*F_active[-1]/kb)
    DT = int(T/10)
    T, DT = str(T), str(DT)
    base, _, nproc, mem, _, _ = busca_input(freqlog)
    print('This is the configuration taken from the file:\n')
    print('Functional/basis: {}'.format(base))
    print('%nproc='+nproc)    
    print('%mem='+mem)
    print('Initial Temperature: {} K'.format(T))
    print('Temperature step: {} K'.format(DT))
    change = input('Are you satisfied with these parameters? y or n?\n')
    if change == 'n':
        base   = default(base,"Functional/basis is {}. If ok, Enter. Otherwise, type functional/basis.\n".format(base))
        nproc  = default(nproc,'nproc={}. If ok, Enter. Otherwise, type it.\n'.format(nproc))
        mem    = default(mem,"mem={}. If ok, Enter. Otherwise, type it.\n".format(mem))
        T      = default(T,"Initial temperature is {} K. If ok, Enter. Otherwise, type it.\n".format(T))
        DT     = default(DT,"Temperature step is {} K. If ok, Enter. Otherwise, type it.\n".format(DT))
    script    = fetch_file('batch script',['.sh'])    
    num_geoms = input("Number of geometries sampled at each round?\n")
    rounds    = input("Number of rounds?\n")
    numjobs   = input("Number of jobs in each batch?\n")
    gaussian  = input('g16 or g09?\n')
    try:
        int(num_geoms)
        int(rounds)
        int(numjobs)
    except:
        fatal_error("These must be integers. Goodbye!")
    with open('limit.lx','w') as f:
        f.write('Running')
    import subprocess
    folder = os.path.dirname(os.path.realpath(__file__)) 
    subprocess.Popen(['nohup', 'python3', folder+'/conf_search.py', freqlog, base, nproc, mem, T, DT, num_geoms, rounds,numjobs, script, gaussian, '&'])
###############################################################


##FINDS SUITABLE VALUE FOR STD#################################    
def detect_sigma():
    try:
        files = [i for i in os.listdir('.') if 'Magnitudes' in i and '.lx' in i]
        file  = files[0]
        temp = float(file.split('_')[1].strip('K'))
        sigma =  np.round(kb*temp,3)
    except:
        sigma = 0.000
    return sigma
###############################################################    

##CHECKS SPECTRUM TYPE#########################################
def get_spec():
    coms = [file for file in os.listdir("Geometries") if 'Geometr' in file and '.com' in file]
    with open('Geometries/'+coms[0],'r') as f:
        for line in f:
            if 'ABSSPCT' in line:
                tipo = 'absorption'
                break
            elif 'EMISPCT' in line:
                tipo = 'emission'
                break
    return tipo       
###############################################################

##FETCHES REFRACTIVE INDEX##################################### 
def get_nr():
    buscar = False
    coms = [file for file in os.listdir("Geometries") if 'Geometr' in file and '.com' in file]
    with open('Geometries/'+coms[0],'r') as f:
        for line in f:
            if 'SCRF' in line.upper():
                buscar = True
                break
    if buscar:
        logs = [file for file in os.listdir("Geometries") if 'Geometr' in file and '.log' in file]
        for log in logs:
            with open('Geometries/'+log,'r') as f:
                for line in f:
                    if 'Solvent' in line and 'Eps' in line:
                        line = line.split()
                        nr = np.sqrt(float(line[6]))
                        return nr
    else:
        return 1                
###############################################################

##FETCHES CHARGE AND MULTIPLICITY##############################
def get_cm(freqlog):
    with open(freqlog,'r') as f:
        for line in f:
            if 'Charge' in line and 'Multiplicity' in line:
                line = line.split()
                charge = line[2]
                mult   = line[5]
                break
    return charge+' '+mult
###############################################################

##QUERY FUNCTION###############################################
def default(a,frase):
    b = input(frase)
    if b == '':
        return a
    else:
        return b    
###############################################################

##SETS DIELECTRIC CONSTANTS####################################
def set_eps(scrf):
    if 'READ' in scrf.upper():
        eps1 = input("Type the static dielectric constant.\n")
        eps2 = input("Type the dynamic dielectric constant (n^2).\n")
        try:
            float(eps1)
            float(eps2)
        except:
            fatal_error("The constants must be numbers. Goodbye!")
        epss = "Eps="+eps1+"\nEpsInf="+eps2+"\n\n"
    else:
        epss = '\n'
    return epss
###############################################################

##STOP SUBMISSION OF JOBS######################################
def abort_batch():
    choice = input('Are you sure you want to prevent new jobs from being submitted? y or n?\n')
    if choice == 'y':
        try:
            os.remove('limit.lx')
            print('Done!')
        except:
            print('Could not find the files. Maybe you are in the wrong folder.')
    else:
        print('OK, nevermind')
###############################################################

##DELETES CHK FILES############################################
def delchk(input,term):
    num = input.split('-')[1]
    if term == 1:
        a = ''
    elif term == 2:
        a = '2'
    try:        
        os.remove('step{}_{}.chk'.format(a,num))
    except:
        pass      
###############################################################

##CHECKS WHETHER JOBS ARE DONE#################################
def watcher(files,counter):
    rodando = files.copy()
    done = []
    for input in rodando: 
        term = 0
        error = False
        try:
            with open(input[:-3]+'log', 'r') as f:
                for line in f:
                    if 'Normal termination' in line:
                        term += 1
                        if counter == 2:
                            delchk(input,term)
                    elif 'Error termination' in line:
                        error = True
                        print('The following job returned an error: {}'.format(input))
                        print('Please check the file for any syntax errors.')        
            if term == counter or error:
                done.append(input)
        except:
            pass 
    for elem in done:
        del rodando[rodando.index(elem)]                                
    return rodando
###############################################################


##GETS SPECTRA#################################################
def search_spectra():
    Abs, Emi = 'None', 'None'
    candidates = [i for i in os.listdir('.') if '.lx' in i]
    for candidate in candidates:
        with open(candidate, 'r') as f:
            for line in f:
                if 'cross_section' in line:
                    Abs = candidate
                elif 'diff_rate' in line:     
                    Emi = candidate
                break
    return Abs, Emi
###############################################################

##RUNS EXCITON ANALYSIS########################################
def ld():
    Abs, Emi = search_spectra()
    print('Absorption file: {}'.format(Abs))
    print('Emission file: {}'.format(Emi))
    check = input('Are these correct? y or n?\n')
    if check == 'n':
        Abs = input('Type name of the absorption spectrum file\n')
        Emi = input('Type name of the emission spectrum file\n')

    kappa = input('Orientation Factor (k^2):\n')
    rmin  = input("Average intermolecular distance in Å:\n")
    Phi   = input("Fluorescence quantum yield (from 0 to 1):\n")
    try:
        rmin  = float(rmin)
        kappa = np.sqrt(float(kappa))
        Phi   = float(Phi) 
    except:
        fatal_error('These features must be numbers. Goodbye!')    
    if Phi > 1 or Phi < 0:
        fatal_error('Quantum yield must be between 0 and 1. Goodbye!')

    correct = input('Include correction for short distances? y or n?\n')
    if correct == 'y':
        alpha = 1.15*0.53 
        print('Employing correction!')
    else:
        alpha = 0
        print('Not employing correction!')
    
    print('Computing...')
    from lx.ld import run_ld 
    try:
        run_ld(Abs, Emi, alpha, rmin, kappa, Phi)
        print('Results can be found in the ld.lx file')
    except:
        print('Something went wrong. Check if the name of the files are correct.')        
###############################################################

##CHECKS WHETHER JOBS ARE DONE#################################
def hold_watch(files, log):
    rodando = files.copy()
    while len(rodando) > 0:
        rodando = watcher(rodando,1)
        if 'limit.lx' not in os.listdir('.'):
            with open(log,'a') as f:
                f.write('\n#Aborted!')
            sys.exit()
        time.sleep(30)    
###############################################################

##RUNS CALCULATIONS############################################
def rodar_lista(lista, batch_file, gaussian, log, num=1): 
    #number of scripts is integer division of number of files by num
    n = len(lista)//num
    for i in range(n):
        with open(f'cmd_{i}.sh', 'w') as f:
            for file in lista[i*num:(i+1)*num]:
                f.write('{} {}\n'.format(gaussian,file))
        subprocess.call(['bash', batch_file, f'cmd_{i}.sh']) 
    if len(lista)%num != 0:
        with open(f'cmd_{n+1}.sh', 'w') as f:
            for file in lista[n*num:]:
                f.write('{} {}\n'.format(gaussian,file))
        subprocess.call(['bash', batch_file, f'cmd_{n+1}.sh'])
    hold_watch(lista, log)
###############################################################



#!/usr/bin/env python3
import sys
import lx.tools

def interface():
    print("#                       #     #")
    print("#        ######   ####   #   # ")
    print("#        #       #    #   # #  ")
    print("#        #####   #    #    #   ")
    print("#        #       #    #   # #  ")
    print("#        #       #    #  #   # ")
    print("#######  ######   ####  #     #")
    print("--SPECTRA - EXCITON - TUNING--\n")
    print("Choose your option:\n")
    print('SPECTRUM SIMULATIONS:')
    print("\t1 - Generate the inputs for the spectrum calculation")
    print("\t2 - Run the spectrum calculations")
    print("\t3 - Generate the spectrum")
    print("\t4 - Check the progress of the calculations")
    print('EXCITON ANALYSIS:')
    print("\t5 - Estimate Förster radius, fluorescence lifetime and exciton diffusion lengths")
    print('CONFORMATIONAL ANALYSIS:')
    print("\t6 - Perform conformational search")
    print('OTHER FEATURES:')
    print("\t7 - Perform long-range parameter tuning") 
    print("\t8 - Retrieve last geometry from log file") 
    print("\t9 - Abort my calculations")
    op = input()
    if op == '1':
        freqlog = lx.tools.fetch_file("frequency",['.log'])
        F, _  = lx.tools.pega_freq(freqlog)
        if F[0] < 0:
            lx.tools.fatal_error("Imaginary frequency! Goodbye!")
        cm = lx.tools.get_cm(freqlog)
        base, temtd, nproc, mem, scrf, spec = lx.tools.busca_input(freqlog)
        if temtd == '':
            tda = 'TD'
        else:
            tda = temtd.upper()
        print('\nThe suggested configurations for you are:\n')
        print('Functional/basis: {}'.format(base))
        print('Solvent Method: {}'.format(scrf))
        print('Charge and Multiplicity: {}'.format(cm))
        print('TD-DFT or TDA-DFT: {}-DFT'.format(tda))
        print('%nproc='+nproc)    
        print('%mem='+mem)
        change = input('Are you satisfied with these parameters? y or n?\n')
        if change.lower() == 'n':     
            base  = lx.tools.default(base,"Functional/basis is {}. If ok, Enter. Otherwise, type functional/basis.\n".format(base))
            scrf  = lx.tools.default(scrf,"SCRF keyword is {}. If ok, Enter. Otherwise, type the desired one.\n".format(scrf))
            cm    = lx.tools.default(cm,'Charge and multiplicity is {}. If ok, Enter. Otherwise, type charge and multiplicity Ex.: 0 1\n'.format(cm))
            nproc = lx.tools.default(nproc,'nproc is {}. If ok, Enter. Otherwise, type it.\n'.format(nproc))
            mem   = lx.tools.default(mem,"mem is {}. If ok, Enter. Otherwise, type it.\n".format(mem))
            tamm  = input('Use TDA (Tamm-Dancoff Approximation)? y or n?\n')
            if tamm.lower() == 'y':
                tda = 'TDA'
            else:
                tda = 'TD'

        num_ex = input("How many excited states?\n")
        try:
            num_ex = int(num_ex)
        except:
            lx.tools.fatal_error("This must be a number! Goodbye!")
        num_geoms = int(input("How many geometries to be sampled?\n"))
        pcm = input("Include state specific solvent approach? y or n?\n")
        if pcm.lower() == 'y':
            solv = input("What is the solvent? If you want to specify the dielectric constants yourself, type read.\n")
            epss = lx.tools.set_eps(solv)
            if epss == '\n':
                solv = "SOLVENT="+solv
            if temtd:
                print("Inputs suitable for emission spectra!\n")    
                header = "%chk=step_UUUUU.chk\n%nproc={}\n%mem={}\n# {} {}=(NSTATES={}) SCRF=(CorrectedLR,NonEquilibrium=Save,{})\n\n{}\n\n{}\n".format(nproc,mem,base,tda,num_ex,solv,spec,cm) 
                bottom = "{}\n--Link1--\n%nproc={}\n%mem={}\n%oldchk=step_UUUUU.chk\n%chk=step2_UUUUU.chk\n# {} GUESS=READ GEOM=CHECKPOINT SCRF(NonEquilibrium=Read,{})\n\nTITLE\n\n{}\n\n{}".format(epss,nproc,mem,base,solv,cm,epss) 
            else:
                print("Inputs suitable for absortion spectra!!\n")
                header = "%nproc={}\n%mem={}\n# {} SCRF=(CorrectedLR,{}) {}=(NSTATES={})\n\n{}\n\n{}\n".format(nproc,mem,base,solv,tda,num_ex,spec,cm)
                bottom = epss
        elif pcm == 'n':
            epss = lx.tools.set_eps(scrf)
            header = "%nproc={}\n%Mem={}\n# {}=(NStates={}) {} {} \n\n{}\n\n{}\n".format(nproc,mem,tda,num_ex,base,scrf,spec,cm)
            bottom = epss+'\n\n'
        else:
            lx.tools.fatal_error("It should be y or n. Goodbye!")
        T = float(input("Temperature in Kelvin?\n"))
        if T <= 0:
            lx.tools.fatal_error("Have you heard about absolute zero? Goodbye!")
        lx.tools.make_ensemble(freqlog, num_geoms, T, header, bottom)    
    elif op == '2':
        lx.tools.batch() 
    elif op == '3':
        opc  = lx.tools.detect_sigma()
        tipo = lx.tools.get_spec()
        nr   = lx.tools.get_nr() 
        print('The spectrum will be run with the following parameters:\n')
        print('Spectrum type: {}'.format(tipo.title()))
        print('Standard deviation of: {:.3f} eV'.format(opc))
        print('Refractive index: {:.3f}\n'.format(nr))
        change = input('Are you satisfied with these parameters? y or n?\n')
        if change.lower() == 'n':
            opc = input("What is the standard deviation of the gaussians?\n")
            try:
                opc = float(opc)
            except: 
                lx.tools.fatal_error("It must be a number. Goodbye!")
            nr = input("What is the refractive index?\n")
            try:
                nr = float(nr)
            except: 
                lx.tools.fatal_error("It must be a number. Goodbye!") 
            tipo = input("What kind of spectrum? Type abs (absorption) or emi (emission)\n")
            if tipo != 'abs' and tipo != 'emi':
                lx.tools.fatal_error('It must be either one. Goodbye!')
        tipo = tipo[:3]
        if tipo == 'abs':
            estados = input("How many excited states?\n")
            try:
                estados = int(estados)
            except:
                lx.tools.fatal_error("It must be an integer! Goodbye!")
        else:
            estados = 1
        num_ex = range(0,estados+1)
        num_ex = list(map(int,num_ex))
        lx.tools.gather_data(opc, tipo)
        lx.tools.spectra(tipo, num_ex, nr)
    elif op == '4':
        lx.tools.andamento()
    elif op == '5':
        lx.tools.ld()
    elif op == '6':
        lx.tools.conformational()
    elif op == '7':
        lx.tools.omega_tuning()
    elif op == '8':
        freqlog = lx.tools.fetch_file("log",['.log'])
        base, _, nproc, mem, scrf, _ = lx.tools.busca_input(freqlog)
        cm = lx.tools.get_cm(freqlog)
        header = '%nproc={}\n%mem={}\n# {} {}\n\nTITLE\n\n{}\n'.format(nproc,mem,base,scrf,cm)
        G, atomos = lx.tools.pega_geom(freqlog)
        lx.tools.write_input(atomos,G,header,'','geom.lx')
        print('Geometry saved in the geom.lx file.')    
    elif op == '9':
        lx.tools.abort_batch()
    else:
        lx.tools.fatal_error("It must be one of the options... Goodbye!")

def main():
    try:
        freqlog = sys.argv[1]
        G, atomos = lx.tools.pega_geom(freqlog)
        print(len(atomos))
        print('\n')    
        for i in range(len(atomos)):
            print("{:2s}  {:.7f}  {:.7f}  {:.7f}".format(atomos[i],G[i,0],G[i,1],G[i,2]))
    except:
        interface()
    
    


    
if __name__ == "__main__":
    sys.exit(main())        




#!/usr/bin/env python3
import numpy as np
from scipy.interpolate import interp1d

##SOME CONSTANTS##############################################
hbar = 6.582119514e-16       #eV s
c    = 299792458             #m/s
pi = np.pi
###############################################################

##CALCULATES FLUORESCENCE LIFETIME IN S########################
def calc_lifetime(xd,yd,dyd):
    #Integrates the emission spectrum
    IntEmi = np.trapz(yd,xd)
    taxa   = (1/hbar)*IntEmi
    error  = (1/hbar)*np.sqrt(np.trapz((dyd**2),xd))
    dlife  = (1/taxa)*(error/taxa)
    return 1/taxa, dlife 
###############################################################

##CALCULATES FORSTER RADIUS####################################
def radius(xa,ya,dya,xd,yd,dyd,kappa):
    #Speed of light
    c = 299792458  #m/s
    
    #Finds the edges of interpolation
    minA = min(xa)
    minD = min(xd)
    maxA = max(xa)
    maxD = max(xd)
    MIN  = max(minA,minD)
    MAX  = min(maxA,maxD)
    
    if MIN > MAX:
        return 0, 0 
    X = np.linspace(MIN, MAX, 1000)
    f1 = interp1d(xa, ya, kind='cubic')
    f2 = interp1d(xd, yd, kind='cubic')
    f3 = interp1d(xa, dya, kind='cubic')
    f4 = interp1d(xd, dyd, kind='cubic')
    

    YA  = f1(X)
    YD  = f2(X)
    DYA = f3(X)
    DYD = f4(X)

    #Calculates the overlap
    Overlap = YA*YD/(X**4)

    #Overlap error
    OverError   = Overlap*np.sqrt((DYA/YA)**2 + (DYD/YD)**2)

    #Integrates overlap
    IntOver = np.trapz(Overlap, X)

    #Integrated Overlap Error
    DeltaOver = np.sqrt(np.trapz((OverError**2),X))       

    #Gets lifetime
    tau, delta_tau = calc_lifetime(xd,yd,dyd)	

    #Calculates radius sixth power
    c *= 1e10
    const   = (hbar**3)*(9*(c**4)*(kappa**2)*tau)/(8*pi)
    radius6 = const*IntOver

    #Relative error in radius6
    delta   = np.sqrt((DeltaOver/IntOver)**2 + (delta_tau/tau)**2)

    #Calculates radius
    radius  = radius6**(1/6)

    #Error in radius
    delta_radius = radius*delta/6

    return radius, delta_radius
###############################################################

##AVG HOPPING DISTANCES########################################
def r_avg(alpha,moment,rmin,dim):
    neighbors = 100
    n1 = neighbors
    n2 = 0
    n3 = 0
    if dim > 1:
        n2 = neighbors
    if dim > 2:
        n3 = neighbors
    numerator = []
    denominator = []
    for i in range(0,n1+1):
        for j in range(0,n2+1):
            for k in range(0,n3+1):
                r = rmin*np.sqrt(i**2+j**2+k**2)
                if r != 0:
                    numerator.append(r/((alpha*moment + r)**6))
                    denominator.append(1.0/((alpha*moment + r)**6))
    average = np.sum(numerator)/np.sum(denominator)
    return average
###############################################################

##1D PROJECTIONS OF DIFFUSION LENGTH###########################
def difflen(radius, alpha, moment, r, dim, phi):
    ld = r*(radius**3)/((alpha*moment+r)**3)
    ld = np.sqrt(phi)*ld/np.sqrt(dim)
    #Conversion to nm
    ld = ld/10
    return ld

######## ANNIHILATION COEFs ###################################
def KSSA(RF,r,tau,error_radius,error_life):
    KSSA_cte         = (4*np.pi/tau)*((RF)**6/(r**3)) # Singlet-Singlet annihilation coef
    RF_6_power_error = np.sqrt(6*((error_radius/RF)**2))
    delta            = np.sqrt( RF_6_power_error**2 + (error_life/tau)**2)
    KSSA_error       = KSSA_cte*delta
    un               = 1E-24 # unit factor 1 AA^3 = 1E-24 cm^3
    KSSA_cte, KSSA_error = KSSA_cte*un, KSSA_error*un     
    return KSSA_cte, KSSA_error
    
def KTTA(RF,r,tau,error_radius,error_life):
    KTTA_cte         = (4*np.pi/tau)*((RF)**6) # Triplet-Triplet annihilation coef
    RF_6_power_error = np.sqrt(6*((error_radius/RF)**2))
    delta            = np.sqrt( RF_6_power_error**2 + (error_life/tau)**2)
    KTTA_error       = KTTA_cte*delta
    un               = 1E-48 # unit factor 1 AA^6 = 1E-48 cm^6
    KTTA_cte, KTTA_error = KTTA_cte*un, KTTA_error*un  
    return KTTA_cte, KTTA_error
############################################################### 

##RETURNS ABS AND EMISSION TYPES###############################   
def emi_abs_types(Abs,Emi):
    abs_type  = 'S0'
    with open(Emi, 'r') as f:
        line = f.readlines()[1].split('->')
        emi_init  = line[0].split()[-1].strip()
        emi_final = line[1].split(':')[0].strip()
    return abs_type,emi_init,emi_final
###############################################################

##Calculates average hopping distances and diffusion length for each case ####   
def get_lds_dists(alpha,moment,rmin,Phi,mean_radius,error_radius):
    # convention: the last list element (with dimension 3) refers to the Amorphous one
    # dists' structure: [average distance,dimension,morphology label] 
    # lds'   structure: [ld , error ld] 
    dists = [[r_avg(alpha,moment,rmin,i),i,"%sD Crystal"%(i)] for i in range(1,4)] #Calculates average hopping distances for each case	
    dists.append([0.25*alpha*moment + 1.25*rmin,3,"Amorphous "])       
    lds = [ [difflen(mean_radius,alpha, moment, dists[i][0], dists[i][1], Phi), 3*difflen(mean_radius,alpha, moment, dists[i][0], dists[i][1], Phi)*(error_radius/mean_radius)] for i in range(len(dists))]
    return lds,dists
###############################################################

def run_ld(Abs, Emi, alpha, rmin, kappa, Phi):
    #Gets the Transition Dipole Moment in atomic units
    try:
        with open(Emi, 'r') as f:
            for line in f:
                moment = float(line.split('=')[1].split()[0])
                break
    except:
        moment = 0

    
    #Loads the data 
    data_abs = np.loadtxt(Abs)
    data_emi = np.loadtxt(Emi)

    #getting the types of abs and emi spectras
    abs_type, emi_init, emi_final  = emi_abs_types(Abs,Emi)

    # template : abs = S0, emi = Sn ---> S0 ===> ('S0','S','S0')
    rates_types = {('S0','S','S0'): 'SSA' , ('T1','T','S0'): 'TTA', ('S0','T','S0'):'Only radius', ('S1','S','S0'):'Only radius'}        
    
    try: 
        setting_spectras = rates_types[(abs_type, emi_init[0], emi_final)]
    except:
        print('The configuration:\nabs %s\nemi %s -> %s\nis not defined! Generating standard calculations anyway ...' %(abs_type,emi_init,emi_final))
        setting_spectras = 'Only radius'
        
    #Gets energies, intensities and errors for the donor and acceptor
    xa, ya, dya = data_abs[:,0], data_abs[:,1], data_abs[:,2]
    xd, yd, dyd = data_emi[:,0], data_emi[:,1], data_emi[:,2]

    #Lifetime calculations
    mean_life, error_life = calc_lifetime(xd,yd,dyd)
    
    #Radius calculations
    mean_radius, error_radius = radius(xa,ya,dya,xd,yd,dyd,kappa)

    #Diffusion length and distance calculations for all conformations
    lds, dists =  get_lds_dists(alpha,moment,rmin,Phi,mean_radius,error_radius)
    
    with open("ld.lx", 'w') as f:
        f.write("Abs: %s, Emi: %s -> %s\n" % (abs_type,emi_init,emi_final))
        f.write("Forster Radius:      {:.1f} +/- {:.1f} AA \n".format(mean_radius,error_radius))
        f.write("Radiative Lifetime:  {:.3e} +/- {:.3e} s\n".format(mean_life,error_life))
        f.write("Avg. Dipole Moment:  {:.1f} a.u. \n".format(moment))
        
        if setting_spectras == "SSA":
            KSSA_cte, KSSA_error = KSSA(mean_radius,rmin,mean_life,error_radius,error_life)
            f.write("Annihilation Coefficient Singlet-Singlet: {:5.2e} +/- {:5.2e} ".format(KSSA_cte, KSSA_error) +" cm^3 s^-1\n")
            f.write("Morphology   Avg_Hop_Distance(AA)  Diffusion_Length(nm)\n")
            [f.write("{:}   {:<19.1f}  {:>.1f} +/- {:>.1f}\n".format(dists[i][2], dists[i][0], lds[i][0], lds[i][1])) for i in range(len(dists))]           
        if setting_spectras == "TTA":
            KTTA_cte, KTTA_error = KTTA(mean_radius,rmin,mean_life,error_radius,error_life)
            f.write("Annihilation Coefficient Triplet-Triplet: {:5.2e} +/- {:5.2e} ".format(KTTA_cte, KTTA_error) +" cm^6 s^-1\n")
        if setting_spectras == "Only radius":
            pass        



#!/usr/bin/env python3
import numpy as np
import os
import sys
import subprocess
import random
import shutil
import lx.tools
from scipy.stats import norm

c     = lx.tools.c
pi    = lx.tools.pi
hbar  = lx.tools.hbar
hbar2 = lx.tools.hbar2
kb    = lx.tools.kb


def distance_matrix(G):
    matrix = np.zeros((1,np.shape(G)[0]))
    for ind in range(np.shape(G)[0]):
            distances = G - G[ind,:]
            distances = np.sqrt(np.sum(np.square(distances),axis=1))    
            matrix = np.vstack((matrix,distances[np.newaxis,:]))
    matrix = matrix[1:,:]
    matrix[matrix==0] = np.inf
    return matrix

def bond(matrix,atoms):
    valence = {1:0.31, 2:0.28, 3:1.28, 4:0.96, 5:0.84, 6:0.76, 7:0.71, 8:0.66, 9:0.57, 10:0.58, 11:1.66, 12:1.41, 13:1.21, 14:1.11,
           15:1.07, 16:1.05, 17:1.02, 18:1.06, 19:2.03, 20:1.76, 21:1.7, 22:1.6, 23:1.53, 24:1.39, 25:1.61, 26:1.52, 27:1.50,
           28:1.24, 29:1.32, 30:1.22, 31:1.22, 32:1.2, 33:1.19, 34:1.20, 35:1.20, 36:1.16}
    CM = np.zeros(np.shape(matrix))
    while np.min(matrix) < 100:
        x,y = np.where(matrix == np.min(matrix))
        x, y = x[0], y[0]         
        if matrix[x,y] < 1.3*(valence[atoms[x]] + valence[atoms[y]]):
            CM[x,y] += 1
            CM[y,x] += 1      
        matrix[x,y] = np.inf
        matrix[y,x] = np.inf   
    return CM      

def fingerprint(file,folder):
    try:
        G, atoms = lx.tools.pega_geom(folder+'/'+file)
        atoms = np.array(atoms).astype(float)
        matrix = distance_matrix(G)
        cm = bond(matrix,atoms)
    except:
        cm = np.zeros((5,5))
    return cm              

##SAMPLES GEOMETRIES###########################################
def make_geoms(freqlog, num_geoms, T, header, bottom):
    lista = []
    counter = lx.tools.start_counter()
    _, atomos, A = lx.tools.sample_geometries(freqlog,num_geoms,T,3000)
    for n in range(0,np.shape(A)[1],3):
        Gfinal = A[:,n:n+3]
        lx.tools.write_input(atomos,Gfinal,header.replace("UUUUU",str((n+3)//3)),bottom.replace("UUUUU",str((n+3)//3)),"Geometry-"+str((n+3)//3+counter)+"-.com")
        lista.append("Geometry-"+str((n+3)//3+counter)+"-.com") 
    return lista      
############################################################### 

##GETS ENERGY FROM THE ORIGINAL FREQ LOG FILE##################
def get_energy_origin(freqlog):
    exc = 0
    with open(freqlog, 'r') as f:
        for line in f:
            if 'SCF Done:' in line:
                line = line.split()
                scf  = float(line[4])*27.2114
            elif 'Total Energy,' in line:
                line = line.split()
                exc  = float(line[4])*27.2114    
            elif 'Rotational constants' in line:
                line = line.split()
                rot  = [float(line[3]),float(line[4]),float(line[5])]    
            elif 'Normal termination' in line:
                if exc != 0:
                    scf = exc
                return scf, np.array([rot[0], rot[1], rot[2]])
###############################################################

##GETS ENERGIES FROM OPT LOG FILES#############################
def get_energies(folder,original_molecule):
    nums,scfs,rots = [], [], []
    files = [i for i in os.listdir(folder) if '.log' in i and 'Geometry' in i]
    for file in files:
        exc = 0
        if np.array_equal(original_molecule, fingerprint(file,folder)):
            with open(folder+'/'+file, 'r') as f:
                num = float(file.split('-')[1])
                for line in f:
                    if 'SCF Done:' in line:
                        line = line.split()
                        scf  = float(line[4])*27.2114
                    elif 'Total Energy,' in line:
                        line = line.split()
                        exc  = float(line[4])*27.2114    
                    elif 'Rotational constants' in line:
                        line = line.split()
                        rot  = [float(line[3]),float(line[4]),float(line[5])]
                    elif 'Normal termination' in line:
                        if exc != 0:
                            scf = exc
                        scfs.append(scf)
                        nums.append(num)
                        rots.append(rot)
                        
    for file in files:
        try:
            shutil.move(file, 'Geometries/'+file)
            shutil.move(file[:-3]+'com', 'Geometries/'+file[:-3]+'com')
        except:
            pass
    return np.array(nums), np.array(scfs), np.array(rots)
###############################################################

def measure(vec1,vec2,cr):
    vec1 = np.array(vec1)     
    vec2 = np.array(vec2)
    cr   = np.array(cr)
    dist = max(abs(vec1 -vec2) - cr)
    distance =  np.heaviside(dist,0)   
    return distance

class Conformation:
    def __init__(self,rot,energy,identity,num) -> None:
        self.rot = rot[np.newaxis,:]
        self.energy = [energy]
        self.identity = identity
        self.std = rot/1000
        self.num = [num]

    def add_rot(self,rot,num,energy):
        self.rot = np.vstack((self.rot,rot[np.newaxis,:]))
        newstd = np.std(self.rot,axis=0)    
        #get higher std
        self.std = np.where(newstd > self.std, newstd, self.std)
        self.num.append(num)
        self.energy.append(energy)

    def merge_rot(self,rot):
        self.rot = np.vstack((self.rot,rot))
        newstd = np.std(self.rot,axis=0)
        #get higher std
        self.std = np.where(newstd > self.std, newstd, self.std)


    def get_avg(self):
        return np.mean(self.rot,axis=0)  



def internal_comparison(conformations):
    remove = []
    for i in range(len(conformations)-1):
        for j in range(i+1,len(conformations)):
            distance = measure(conformations[i].get_avg(),conformations[j].get_avg(),conformations[j].std+conformations[i].std)
            if distance == 0:
                conformations[i].num += conformations[j].num
                conformations[i].energy += conformations[j].energy
                conformations[i].merge_rot(conformations[j].rot)
                remove.append(j)
                break
    # remove elements from list whose index is in remove
    conformations = [i for j, i in enumerate(conformations) if j not in remove]
    return conformations        
     

def classify(conformations,folder):
    nums, scfs, rots = get_energies(folder,conformations[0].identity)
    for i in range(rots.shape[0]):
        for conformation in conformations:
            distance = measure(rots[i,:],conformation.get_avg(),conformation.std)
            if distance == 0:
                conformation.add_rot(rots[i,:],nums[i],scfs[i])
                break
            conformations.append(Conformation(rots[i,:],scfs[i],conformations[0].identity,nums[i])) 
    if len(conformations) > 1:
        conformations  = internal_comparison(conformations)           
    return conformations

def write_report(conformations,round,total_rounds,temp):
    engs = []
    for conformation in conformations:
        engs.append(np.mean(conformation.energy))
    engs = np.array(engs)    
    argsort = np.argsort(engs)
    engs = engs[argsort]
    #sort conformations as list
    conformations = [conformations[i] for i in argsort]
    deltae = engs - min(engs)
    probs = np.exp(-(deltae)/(0.026))
    probs = probs/sum(probs)

    with open('conformation.lx','w') as f:
        f.write('{:6}  {:10}  {:10}  {:12}  {:10}  {:10}  {:10}  {:10}  {:10}  {:10}  {:6}  {:6}\n'.format('#Group','Energy(eV)','DeltaE(eV)','Prob@300K(%)','Rot1','Rot2','Rot3','Std1','Std2','Std3','Number','Last'))
        for i in range(len(conformations)):
            rot = conformations[i].get_avg()
            std = conformations[i].std
            last = conformations[i].num[-1]
            total= len(conformations[i].num)
            f.write(f'{i+1:<6}  {engs[i]:<10.3f}  {deltae[i]:<10.3f}  {100*probs[i]:<12.1f}  {rot[0]:<10.7f}  {rot[1]:<10.7f}  {rot[2]:<10.7f}  {std[0]:<10.7f}  {std[1]:<10.7f}  {std[2]:<10.7f}  {total:<6.0f}  {last:<6.0f}\n')
        f.write(f'\n#Round {round}/{total_rounds} Temperature: {temp} K')    


##RUNS FREQ CALCULATION FOR NEW CONFORMATION###################
def rodar_freq(origin,nproc,mem,base,cm,batch_file,gaussian):
    geomlog = f'Geometries/Geometry-{origin:.0f}-.log'
    G, atomos = lx.tools.pega_geom(geomlog) 
    header = "%nproc={}\n%mem={}\n# freq=(noraman) nosymm  {} \n\n{}\n\n{}\n".format(nproc,mem,base,'ABSSPCT',cm)
    file = f"Freq-{origin:.0f}-.com"
    lx.tools.write_input(atomos,G,header,'',file)
    lx.tools.rodar_lista([file],batch_file,gaussian,'conformation.lx',1)
    log = file[:-3]+'log'
    with open(log, 'r') as f:
        for line in f:
            if 'Normal termination' in line:
                return log
            elif 'Error termination' in line:
                return None
###############################################################

def main():
    freqlog   = sys.argv[1]
    base      = sys.argv[2]
    nproc     = sys.argv[3]
    mem       = sys.argv[4]
    T         = float(sys.argv[5])
    DT        = float(sys.argv[6])
    num_geoms = int(sys.argv[7])
    rounds    = int(sys.argv[8])
    numjobs   = int(sys.argv[9])
    script    = sys.argv[10]
    gaussian  = sys.argv[11]
    T0 = T
    freq0 = freqlog
    if 'td' in base.lower():
        opt = '=loose'
    else:
        opt = ''    
    
    try:
        os.mkdir('Geometries')
    except:
        pass    
    cm        = lx.tools.get_cm(freqlog) 
    header    = f"%nproc={nproc}\n%mem={mem}\n# opt{opt} nosymm  {base} \n\nTitle\n\n{cm}\n"
    scf, rot = get_energy_origin(freqlog)
    conformations = [Conformation(rot,scf,fingerprint(freqlog,'.'),0)]
    files = [i for i in os.listdir('Geometries') if 'Geometry' in i and '.log' in i]
    if len(files) > 0:
        conformations = classify(conformations,'Geometries')
        write_report(conformations,0,rounds,T0)
    else:
        pass    

    
    groups = len(conformations)
    for i in range(rounds):
        lista      = make_geoms(freqlog, num_geoms, T0, header, '')
        lx.tools.rodar_lista(lista,script, gaussian, 'conformation.lx',numjobs)
        conformations  = classify(conformations,'.')
        write_report(conformations,i+1,rounds,T0)
        
        if len(conformations) != groups:
            log  = rodar_freq(conformations[-1].num[-1],nproc,mem,base,cm,script,gaussian)
            if log != None:
                freqlog = log
                T0 = T
            groups = len(conformations)    
        else:
            T0 += DT

    with open('conformation.lx', 'a') as f:
        f.write('\n#Search concluded!')

    try:
        os.mkdir('Conformers')
    except:
        pass    

    for i in range(len(conformations)):
        numero  = conformations[i].num[-1]
        if numero == 0:
            freqlog = freq0    
        else:
            freqlog = 'Geometries/Geometry-{:.0f}-.log'.format(numero) 
        _, _, nproc, mem, scrf, _ = lx.tools.busca_input(freqlog)
        cm = lx.tools.get_cm(freqlog)
        header = '%nproc={}\n%mem={}\n%chk=Group_{}_.chk\n# {} {} opt\n\nTITLE\n\n{}\n'.format(nproc,mem,i+1,'pm6',scrf,cm)
        G, atomos = lx.tools.pega_geom(freqlog)
        lx.tools.write_input(atomos,G,header,'','Conformers/Geometry-{}-.com'.format(i+1))
    



if __name__ == "__main__":
    sys.exit(main())        


        
        
        


    
        

